@@ -13,7 +13,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
+ * MA  02110-1301, USA.
  */
 
 #include "config.h"
@@ -21,6 +22,7 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <grp.h>
@@ -60,6 +62,11 @@
 #define debug_msg(...) ((void)0)
 #endif
 
+/* Execute permissions. */
+#ifdef HAVE_FEXECVE
+#  define UH_S_IXALL (S_IXUSR | S_IXGRP | S_IXOTH)
+#endif
+
 /* A maximum GECOS field length.  There's no hard limit, so we guess. */
 #define GECOS_LENGTH			127
 
@@ -141,7 +148,7 @@ die(struct app_data *data, int status)
 
 /* Exit, returning the proper status code based on a PAM error code.
    Optionally write a message to stderr. */
-static int G_GNUC_NORETURN
+static void G_GNUC_NORETURN
 fail_exit(struct app_data *data, int pam_retval)
 {
 	int status;
@@ -582,7 +589,7 @@ converse_console(int num_msg, const struct pam_message **msg,
 
 	if (banner == 0) {
 		if ((data->banner != NULL) && (data->domain != NULL)) {
-			text = g_strdup_printf(dgettext(data->domain, data->banner));
+			text = g_strdup (dgettext(data->domain, data->banner));
 		} else {
 			if ((service != NULL) && (strlen(service) > 0)) {
 				if (data->fallback_allowed) {
@@ -750,6 +757,10 @@ pipe_conv_wait_for_sync(struct app_data *data)
 	char *reply;
 	int err;
 
+	if (data->input == NULL) {
+		err = ERR_UNK_ERROR;
+		goto err;
+	}
 	reply = read_reply(data->input);
 	if (reply == NULL) {
 		err = ERR_UNK_ERROR;
@@ -790,6 +801,8 @@ pipe_conv_exec_start(struct app_data *data)
 	if (data->conv->conv == converse_pipe) {
 		/* There might have been no converse_pipe() call, so send the
 		   information now. */
+		if (data->output == NULL)
+			return ERR_UNK_ERROR;
 		send_sn_info(data);
 		send_request(data->output, UH_EXEC_START, NULL);
 		send_request(data->output, UH_SYNC_POINT, NULL);
@@ -1199,6 +1212,25 @@ passwd(const char *user, struct app_data *data)
 	exit(0);
 }
 
+/* Does STRING contain control characters or characters from FORBIDDEN? */
+static gboolean
+string_has_forbidden_characters(const char *string, const char *forbidden)
+{
+	char c;
+
+	if (strpbrk(string, forbidden) != NULL)
+		return TRUE;
+	for (; (c = *string) != 0; string++) {
+		/* This is only a minimal sanity check, primarily to be
+		 * consistent with util-linux, not really a security
+		 * boundary. */
+		if (iscntrl((unsigned char)c))
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
 /* We're here to change the user's non-security information.  PAM doesn't
  * provide an interface to do this, because it's not PAM's job to manage this
  * stuff, so farm it out to a different library. */
@@ -1214,8 +1246,6 @@ chfn(const char *user, struct app_data *data, lu_prompt_fn *prompt,
 	struct lu_context *context;
 	struct lu_ent *ent;
 	struct lu_error *error;
-	GValueArray *values;
-	GValue *value, val;
 	int tryagain = 3, retval;
 	gboolean ret;
 
@@ -1229,13 +1259,20 @@ chfn(const char *user, struct app_data *data, lu_prompt_fn *prompt,
 
 	/* Verify that the fields we were given on the command-line
 	 * are sane (i.e., contain no forbidden characters). */
-	if (new_full_name && strpbrk(new_full_name, ":,="))
+	if (new_full_name != NULL &&
+	    string_has_forbidden_characters(new_full_name, ":,=\n"))
 		die(data, ERR_FIELDS_INVALID);
-	if (new_office && strpbrk(new_office, ":,="))
+	if (new_office != NULL &&
+	    string_has_forbidden_characters(new_office, ":,=\n"))
 		die(data, ERR_FIELDS_INVALID);
-	if (new_office_phone && strpbrk(new_office_phone, ":,="))
+	if (new_office_phone != NULL &&
+	    string_has_forbidden_characters(new_office_phone, ":,=\n"))
 		die(data, ERR_FIELDS_INVALID);
-	if (new_home_phone && strpbrk(new_home_phone, ":,="))
+	if (new_home_phone != NULL &&
+	    string_has_forbidden_characters(new_home_phone, ":,=\n"))
+		die(data, ERR_FIELDS_INVALID);
+	if (new_shell != NULL &&
+	    string_has_forbidden_characters(new_shell, ":\n"))
 		die(data, ERR_FIELDS_INVALID);
 
 	/* Start up PAM to authenticate the user, this time pretending
@@ -1324,10 +1361,8 @@ chfn(const char *user, struct app_data *data, lu_prompt_fn *prompt,
 
 	/* Pull up the user's GECOS data, and split it up. */
 	memset(&parsed_gecos, 0, sizeof(parsed_gecos));
-	values = lu_ent_get(ent, LU_GECOS);
-	if (values != NULL) {
-		value = g_value_array_get_nth(values, 0);
-		old_gecos = lu_value_strdup(value);
+	old_gecos = lu_ent_get_first_value_strdup(ent, LU_GECOS);
+	if (old_gecos != NULL) {
 		gecos_parse(old_gecos, &parsed_gecos);
 		debug_msg("userhelper: old gecos string \"%s\"\n", old_gecos);
 		debug_msg("userhelper: old gecos \"'%s','%s','%s','%s','%s'\""
@@ -1364,7 +1399,7 @@ chfn(const char *user, struct app_data *data, lu_prompt_fn *prompt,
 
 	/* Verify that the strings we got passed are not too long. */
 	if (gecos_size(&parsed_gecos) > GECOS_LENGTH) {
-		debug_msg("userhelper: user gecos too long %d > %d\n",
+		debug_msg("userhelper: user gecos too long %zu > %d\n",
 			  gecos_size(&parsed_gecos), GECOS_LENGTH);
 		lu_ent_free(ent);
 		lu_end(context);
@@ -1376,33 +1411,23 @@ chfn(const char *user, struct app_data *data, lu_prompt_fn *prompt,
 	new_gecos = gecos_assemble(&parsed_gecos);
 	debug_msg("userhelper: new gecos string \"%s\"\n", new_gecos);
 
-	/* We don't need the user's current GECOS anymore, so clear
-	 * out the value and set our own in the in-memory structure. */
-	memset(&val, 0, sizeof(val));
-	g_value_init(&val, G_TYPE_STRING);
-
-	lu_ent_clear(ent, LU_GECOS);
-	g_value_set_string(&val, new_gecos);
-	lu_ent_add(ent, LU_GECOS, &val);
+	/* We don't need the user's current GECOS anymore, so set our own in
+	   the in-memory structure. */
+	lu_ent_set_string(ent, LU_GECOS, new_gecos);
 	g_free(new_gecos);
 
-	/* While we're at it, set the individual data items as well. */
-	lu_ent_clear(ent, LU_COMMONNAME);
-	g_value_set_string(&val, parsed_gecos.full_name);
-	lu_ent_add(ent, LU_COMMONNAME, &val);
-
-	lu_ent_clear(ent, LU_ROOMNUMBER);
-	g_value_set_string(&val, parsed_gecos.office);
-	lu_ent_add(ent, LU_ROOMNUMBER, &val);
-
-	lu_ent_clear(ent, LU_TELEPHONENUMBER);
-	g_value_set_string(&val, parsed_gecos.office_phone);
-	lu_ent_add(ent, LU_TELEPHONENUMBER, &val);
-
-	lu_ent_clear(ent, LU_HOMEPHONE);
-	g_value_set_string(&val, parsed_gecos.home_phone);
-	lu_ent_add(ent, LU_HOMEPHONE, &val);
-
+	/* While we're at it, set the individual data items as well. Note that
+	 * this may modify old data if LU_GECOS and the individual components
+	 * weren’t in sync. */
+	if (parsed_gecos.full_name != NULL)
+		lu_ent_set_string(ent, LU_COMMONNAME, parsed_gecos.full_name);
+	if (parsed_gecos.office != NULL)
+		lu_ent_set_string(ent, LU_ROOMNUMBER, parsed_gecos.office);
+	if (parsed_gecos.office_phone != NULL)
+		lu_ent_set_string(ent, LU_TELEPHONENUMBER,
+				  parsed_gecos.office_phone);
+	if (parsed_gecos.home_phone != NULL)
+		lu_ent_set_string(ent, LU_HOMEPHONE, parsed_gecos.home_phone);
 	gecos_free(&parsed_gecos);
 
 	/* If we're here to change the user's shell, too, do that while we're
@@ -1411,13 +1436,9 @@ chfn(const char *user, struct app_data *data, lu_prompt_fn *prompt,
 	if (new_shell != NULL) {
 		/* Check that the user's current shell is valid, and that she
 		 * is not attempting to change to an invalid shell. */
-		values = lu_ent_get(ent, LU_LOGINSHELL);
-		if (values != NULL) {
-			value = g_value_array_get_nth(values, 0);
-			old_shell = lu_value_strdup(value);
-		} else {
+		old_shell = lu_ent_get_first_value_strdup(ent, LU_LOGINSHELL);
+		if (old_shell == NULL)
 			old_shell = g_strdup("/bin/sh");
-		}
 
 		debug_msg("userhelper: current shell \"%s\"\n", old_shell);
 		debug_msg("userhelper: new shell \"%s\"\n", new_shell);
@@ -1432,9 +1453,7 @@ chfn(const char *user, struct app_data *data, lu_prompt_fn *prompt,
 		}
 
 		/* Set the shell to the new value. */
-		lu_ent_clear(ent, LU_LOGINSHELL);
-		g_value_set_string(&val, new_shell);
-		lu_ent_add(ent, LU_LOGINSHELL, &val);
+		lu_ent_set_string(ent, LU_LOGINSHELL, new_shell);
 	}
 
 	/* Save the changes to the user's account to the password
@@ -1489,15 +1508,18 @@ wrap(const char *user, const char *program,
 	 * user given in the console.apps configuration file, execute the
 	 * command given in the console.apps file. */
 	char *constructed_path;
+#ifdef HAVE_FEXECVE
+	int fd;
+#endif
 	char *apps_filename;
 	char *user_pam;
 	const char *auth_user;
 	char *val;
 	char **environ_save, **keep_env_names, **keep_env_values;
-	const char *env_home, *env_term, *env_desktop_startup_id;
-	const char *env_display, *env_shell;
-	const char *env_lang, *env_language, *env_lcall, *env_lcmsgs;
-	const char *env_xauthority;
+	char *env_home, *env_term, *env_desktop_startup_id;
+	char *env_display, *env_shell;
+	char *env_lang, *env_language, *env_lcall, *env_lcmsgs;
+	char *env_xauthority;
 	int session, tryagain, gui, retval;
 	struct stat sbuf;
 	struct passwd *pwd;
@@ -1530,16 +1552,16 @@ wrap(const char *user, const char *program,
 
 	/* Save some of the current environment variables, because the
 	 * environment is going to be nuked shortly. */
-	env_desktop_startup_id = getenv("DESKTOP_STARTUP_ID");
-	env_display = getenv("DISPLAY");
-	env_home = getenv("HOME");
-	env_lang = getenv("LANG");
-	env_language = getenv("LANGUAGE");
-	env_lcall = getenv("LC_ALL");
-	env_lcmsgs = getenv("LC_MESSAGES");
-	env_shell = getenv("SHELL");
-	env_term = getenv("TERM");
-	env_xauthority = getenv("XAUTHORITY");
+	env_desktop_startup_id = g_strdup(getenv("DESKTOP_STARTUP_ID"));
+	env_display = g_strdup(getenv("DISPLAY"));
+	env_home = g_strdup(getenv("HOME"));
+	env_lang = g_strdup(getenv("LANG"));
+	env_language = g_strdup(getenv("LANGUAGE"));
+	env_lcall = g_strdup(getenv("LC_ALL"));
+	env_lcmsgs = g_strdup(getenv("LC_MESSAGES"));
+	env_shell = g_strdup(getenv("SHELL"));
+	env_term = g_strdup(getenv("TERM"));
+	env_xauthority = g_strdup(getenv("XAUTHORITY"));
 
 	/* Sanity-check the environment variables as best we can: those
 	 * which aren't path names shouldn't contain "/", and none of
@@ -1579,28 +1601,36 @@ wrap(const char *user, const char *program,
 	if (env_term &&
 	    (strstr(env_term, "..") ||
 	     strchr(env_term, '%')))
-		env_term = "dumb";
+		env_term = g_strdup("dumb");
 	if (env_xauthority &&
 	    (strstr(env_xauthority , "..") ||
 	     strchr(env_xauthority , '%')))
 		env_xauthority = NULL;
 
+	keep_env_names = NULL;
+	keep_env_values = NULL;
 	val = svGetValue(s, "KEEP_ENV_VARS");
 	if (val != NULL) {
 		size_t i, num_names;
 
 		keep_env_names = g_strsplit(val, ",", -1);
 		g_free(val);
-		num_names = g_strv_length(keep_env_names);
-		keep_env_values = g_malloc0_n(num_names,
+		if (keep_env_names) {
+			num_names = g_strv_length(keep_env_names);
+			keep_env_values = g_malloc0_n(num_names,
 					      sizeof (*keep_env_values));
-		for (i = 0; i < num_names; i++)
-			/* g_strdup(NULL) is defined to be NULL. */
-			keep_env_values[i]
-				= g_strdup(getenv(keep_env_names[i]));
-	} else {
-		keep_env_names = NULL;
-		keep_env_values = NULL;
+			if (keep_env_values)
+				for (i = 0; i < num_names; i++)
+				/* g_strdup(NULL) is defined to be NULL. */
+				keep_env_values[i]
+					= g_strdup(getenv(keep_env_names[i]));
+		}
+		if (keep_env_names == NULL || keep_env_values == NULL) {
+			if (keep_env_names) g_strfreev(keep_env_names);
+			if (keep_env_values) g_strfreev(keep_env_values);
+			keep_env_names = NULL;
+			keep_env_values = NULL;
+		}
 	}
 
 	/* Wipe out the current environment. */
@@ -1612,17 +1642,40 @@ wrap(const char *user, const char *program,
 	 * get at others' X authority records -- we restore XAUTHORITY below
 	 * *after* successfully authenticating, or abandoning authentication in
 	 * order to run the wrapped program as the invoking user. */
-	if (env_display) setenv("DISPLAY", env_display, 1);
+	if (env_display) {
+		setenv("DISPLAY", env_display, 1);
+		g_free(env_display);
+	}
 
 	/* The rest of the environment variables are simpler. */
-	if (env_desktop_startup_id) setenv("DESKTOP_STARTUP_ID",
-					   env_desktop_startup_id, 1);
-	if (env_lang) setenv("LANG", env_lang, 1);
-	if (env_language) setenv("LANGUAGE", env_language, 1);
-	if (env_lcall) setenv("LC_ALL", env_lcall, 1);
-	if (env_lcmsgs) setenv("LC_MESSAGES", env_lcmsgs, 1);
-	if (env_shell) setenv("SHELL", env_shell, 1);
-	if (env_term) setenv("TERM", env_term, 1);
+	if (env_desktop_startup_id) {
+		setenv("DESKTOP_STARTUP_ID", env_desktop_startup_id, 1);
+		g_free(env_desktop_startup_id);
+	}
+	if (env_lang) {
+		setenv("LANG", env_lang, 1);
+		g_free(env_lang);
+	}
+	if (env_language) {
+		setenv("LANGUAGE", env_language, 1);
+		g_free(env_language);
+	}
+	if (env_lcall) {
+		setenv("LC_ALL", env_lcall, 1);
+		g_free(env_lcall);
+	}
+	if (env_lcmsgs) {
+		setenv("LC_MESSAGES", env_lcmsgs, 1);
+		g_free(env_lcmsgs);
+	}
+	if (env_shell) {
+		setenv("SHELL", env_shell, 1);
+		g_free(env_shell);
+	}
+	if (env_term) {
+		setenv("TERM", env_term, 1);
+		g_free(env_term);
+	}
 
 	/* Set the PATH to a reasonaly safe list of directories. */
 	setenv("PATH", "/usr/sbin:/usr/bin:/sbin:/bin:/root/bin", 1);
@@ -1677,14 +1730,26 @@ wrap(const char *user, const char *program,
 		   off the parent to a "pass exit code through" mode without
 		   displaying any unwanted GUI dialogs. */
 		retval = pipe_conv_exec_start(data);
-		if (retval != 0)
+		if (retval != 0) {
+			g_strfreev(environ);
+			environ = environ_save;
 			die(data, retval);
+		}
 		data->conv = text_conv;
 	}
 
 	/* Read the path to the program to run. */
 	constructed_path = svGetValue(s, "PROGRAM");
+	/* Prefer fexecve to prevent race contitions. */
+#ifdef HAVE_FEXECVE
+	fd = constructed_path ? open(constructed_path, O_RDONLY) : -1;
+	if (!constructed_path || fd < 0 || constructed_path[0] != '/') {
+		struct stat statbuf;
+		
+		if (fd >= 0) close(fd);
+#else
 	if (!constructed_path || constructed_path[0] != '/') {
+#endif
 		g_free(constructed_path);
 		/* Criminy....  The system administrator didn't give us an
 		 * absolute path to the program!  Guess either /usr/sbin or
@@ -1693,18 +1758,57 @@ wrap(const char *user, const char *program,
 		 * app, so access() may not be correct here, as it may give
 		 * false negatives.  But then, it wasn't an absolute path. */
 		constructed_path = g_strconcat("/usr/sbin/", program, NULL);
+#ifdef HAVE_FEXECVE
+		if (
+			(fd = open(constructed_path, O_RDONLY)) < 0
+		    || fstat(fd, &statbuf) < 0
+		    || (statbuf.st_mode & UH_S_IXALL) != UH_S_IXALL
+		) {
+			if (fd >= 0) close(fd);
+#else
 		if (access(constructed_path, X_OK) != 0) {
+#endif
 			/* Try the second directory. */
 			strcpy(constructed_path, "/sbin/");
 			strcat(constructed_path, program);
+#ifdef HAVE_FEXECVE
+			if (
+				(fd = open(constructed_path, O_RDONLY)) < 0
+				|| fstat(fd, &statbuf) < 0
+				|| (statbuf.st_mode & UH_S_IXALL) != UH_S_IXALL
+			) {
+				if (fd >= 0) close(fd);
+#else
 			if (access(constructed_path, X_OK)) {
+#endif
 				/* Nope, not there, either. */
 				debug_msg("userhelper: couldn't find wrapped "
 					  "binary\n");
+				g_strfreev(environ);
+				environ = environ_save;
 				die(data, ERR_NO_PROGRAM);
 			}
 		}
 	}
+#ifdef HAVE_FEXECVE
+    {
+		char buf[2];
+		ssize_t nread;
+		int flags;
+
+        flags = fcntl(fd, F_GETFD);
+        if (flags < 0) {
+			g_warning("fcntl(%d, F_GETFD) < 0", fd);
+			flags = 0;
+		}
+		nread = read(fd, buf, sizeof(buf));
+		lseek(fd, 0, SEEK_SET);
+		/* fexecve(2): if executable is a script, don't close on exec */
+		if (nread < (ssize_t)sizeof(buf) || buf[0] != '#' || buf[1] != '!')
+			if (fcntl(fd, F_SETFD, flags | FD_CLOEXEC) < 0)
+				g_warning("fcntl(%d, F_SETFD, FD_CLOEXEC) < 0", fd);
+	}
+#endif
 
 	user_pam = get_user_for_auth(data, s);
 	/* Verify that the user we need to authenticate as has a home
@@ -1712,6 +1816,11 @@ wrap(const char *user, const char *program,
 	pwd = getpwnam(user_pam);
 	if (pwd == NULL) {
 		debug_msg("userhelper: no user named %s exists\n", user_pam);
+#ifdef HAVE_FEXECVE
+		close(fd);
+#endif
+		g_strfreev(environ);
+		environ = environ_save;
 		die(data, ERR_NO_USER);
 	}
 
@@ -1722,8 +1831,11 @@ wrap(const char *user, const char *program,
 	else {
 		/* Otherwise, if they had a reasonable value for HOME, let them
 		 * use it. */
-		if (env_home != NULL)
+		if (env_home != NULL) {
 			setenv("HOME", env_home, 1);
+			g_free(env_home);
+			env_home = NULL;
+		}
 		else {
 			/* Otherwise, set HOME to the user's home directory. */
 			pwd = getpwuid(getuid());
@@ -1731,6 +1843,7 @@ wrap(const char *user, const char *program,
 				setenv("HOME", pwd->pw_dir, 1);
 		}
 	}
+	if (env_home) g_free(env_home);
 
 	/* Read other settings. */
 	session = svTrueValue(s, "SESSION", FALSE);
@@ -1784,6 +1897,11 @@ wrap(const char *user, const char *program,
 	retval = pam_start(program, user_pam, data->conv, &data->pamh);
 	if (retval != PAM_SUCCESS) {
 		debug_msg("userhelper: pam_start() failed\n");
+#ifdef HAVE_FEXECVE
+		close(fd);
+#endif
+		g_strfreev(environ);
+		environ = environ_save;
 		fail_exit(data, retval);
 	}
 
@@ -1802,25 +1920,37 @@ wrap(const char *user, const char *program,
 	if (retval != PAM_SUCCESS) {
 		pam_end(data->pamh, retval);
 		if (data->canceled) {
+#ifdef HAVE_FEXECVE
+			close(fd);
+#endif
+			g_strfreev(environ);
+			environ = environ_save;
 			fail_exit(data, retval);
 		} else
 		if (data->fallback_allowed) {
 			/* Reset the user's environment so that the
 			 * application can run normally. */
 			argv[optind - 1] = strdup(program);
+			g_strfreev(environ);
 			environ = environ_save;
 			become_normal(data, user);
 			if (data->input != NULL) {
 				fflush(data->input);
-				fcntl(UH_INFILENO, F_SETFD, FD_CLOEXEC);
+				if (fcntl(UH_INFILENO, F_SETFD, FD_CLOEXEC) < 0)
+					g_warning("fcntl(%d, F_SETFD, FD_CLOEXEC) < 0", UH_INFILENO);
 			}
 			if (data->output != NULL) {
 				fflush(data->output);
-				fcntl(UH_OUTFILENO, F_SETFD, FD_CLOEXEC);
+				if (fcntl(UH_OUTFILENO, F_SETFD, FD_CLOEXEC) < 0)
+					g_warning("fcntl(%d, F_SETFD, FD_CLOEXEC) < 0", UH_OUTFILENO);
 			}
 			retval = pipe_conv_exec_start(data);
-			if (retval != 0)
+			if (retval != 0) {
+#ifdef HAVE_FEXECVE
+				close(fd);
+#endif
 				die(data, retval);
+			}
 #ifdef USE_STARTUP_NOTIFICATION
 			if (data->sn_id) {
 				debug_msg("userhelper: setting "
@@ -1830,11 +1960,22 @@ wrap(const char *user, const char *program,
 				       data->sn_id, 1);
 			}
 #endif
+#ifdef USE_FEXECVE
+			fexecve(fd, argv + optind - 1, environ);
+			close(fd);
+#else
 			execv(constructed_path, argv + optind - 1);
-			pipe_conv_exec_fail(data);
+#endif
+			if (data->output != NULL)
+				pipe_conv_exec_fail(data);
 			die(data, ERR_EXEC_FAILED);
 		} else {
 			/* Well, we tried. */
+#ifdef HAVE_FEXECVE
+			close(fd);
+#endif
+			g_strfreev(environ);
+			environ = environ_save;
 			fail_exit(data, retval);
 		}
 	}
@@ -1843,17 +1984,33 @@ wrap(const char *user, const char *program,
 	 * out trying to authenticate. */
 	retval = get_pam_string_item(data->pamh, PAM_USER, &auth_user);
 	if (retval != PAM_SUCCESS) {
+#ifdef HAVE_FEXECVE
+		close(fd);
+#endif
 		pam_end(data->pamh, retval);
+		g_strfreev(environ);
+		environ = environ_save;
 		fail_exit(data, retval);
 	}
-	if (strcmp(user_pam, auth_user) != 0)
+	if (strcmp(user_pam, auth_user) != 0) {
+#ifdef HAVE_FEXECVE
+		close(fd);
+#endif
+		g_strfreev(environ);
+		environ = environ_save;
 		die(data, ERR_UNK_ERROR);
+	}
 
 	/* Verify that the authenticated user is allowed to run this
 	 * service now. */
 	retval = pam_acct_mgmt(data->pamh, 0);
 	if (retval != PAM_SUCCESS) {
+#ifdef HAVE_FEXECVE
+		close(fd);
+#endif
 		pam_end(data->pamh, retval);
+		g_strfreev(environ);
+		environ = environ_save;
 		fail_exit(data, retval);
 	}
 
@@ -1862,6 +2019,11 @@ wrap(const char *user, const char *program,
 	pwd = getpwnam(user_pam);
 	if (pwd == NULL) {
 		debug_msg("userhelper: no user named %s exists\n", user_pam);
+#ifdef HAVE_FEXECVE
+		close(fd);
+#endif
+		g_strfreev(environ);
+		environ = environ_save;
 		die(data, ERR_NO_USER);
 	}
 
@@ -1875,12 +2037,18 @@ wrap(const char *user, const char *program,
 		 * environment variable. */
 		if (env_xauthority) {
 			setenv("XAUTHORITY", env_xauthority, 1);
+			g_free(env_xauthority);
 		}
 
 		/* Open a session. */
 		retval = pam_open_session(data->pamh, 0);
 		if (retval != PAM_SUCCESS) {
+#ifdef HAVE_FEXECVE
+			close(fd);
+#endif
 			pam_end(data->pamh, retval);
+			g_strfreev(environ);
+			environ = environ_save;
 			fail_exit(data, retval);
 		}
 
@@ -1888,14 +2056,25 @@ wrap(const char *user, const char *program,
 
 		retval = pam_setcred(data->pamh, PAM_ESTABLISH_CRED);
 		if (retval != PAM_SUCCESS) {
+#ifdef HAVE_FEXECVE
+			close(fd);
+#endif
 			pam_end(data->pamh, retval);
+			g_strfreev(environ);
+			environ = environ_save;
 			fail_exit(data, retval);
 		}
 
 		/* Start up a child process we can wait on. */
 		child = fork();
-		if (child == -1)
+		if (child == -1) {
+#ifdef HAVE_FEXECVE
+			close(fd);
+#endif
+			g_strfreev(environ);
+			environ = environ_save;
 			die(data, ERR_EXEC_FAILED);
+		}
 		if (child == 0) {
 			/* We're in the child.  Make a few last-minute
 			 * preparations and exec the program. */
@@ -1917,15 +2096,23 @@ wrap(const char *user, const char *program,
 			become_super_other(data);
 			if (data->input != NULL) {
 				fflush(data->input);
-				fcntl(UH_INFILENO, F_SETFD, FD_CLOEXEC);
+				if (fcntl(UH_INFILENO, F_SETFD, FD_CLOEXEC) < 0)
+					g_warning("fcntl(%d, F_SETFD, FD_CLOEXEC) < 0", UH_INFILENO);
 			}
 			if (data->output != NULL) {
 				fflush(data->output);
-				fcntl(UH_OUTFILENO, F_SETFD, FD_CLOEXEC);
+				if (fcntl(UH_OUTFILENO, F_SETFD, FD_CLOEXEC) < 0)
+					g_warning("fcntl(%d, F_SETFD, FD_CLOEXEC) < 0", UH_OUTFILENO);
 			}
 			retval = pipe_conv_exec_start(data);
-			if (retval != 0)
+			if (retval != 0) {
+#ifdef HAVE_FEXECVE
+				close(fd);
+#endif
+				g_strfreev(environ);
+				environ = environ_save;
 				die(data, retval);
+			}
 #ifdef USE_STARTUP_NOTIFICATION
 			if (data->sn_id) {
 				debug_msg("userhelper: setting "
@@ -1943,11 +2130,19 @@ wrap(const char *user, const char *program,
 			syslog(LOG_NOTICE, "running '%s' with "
 			       "root privileges on behalf of '%s'",
 			       cmdline, user);
+#ifdef HAVE_FEXECVE
+			fexecve(fd, argv + optind - 1, environ);
+			close(fd);
+#else
 			execv(constructed_path, argv + optind - 1);
+#endif
 			syslog(LOG_ERR, "could not run '%s' with "
 			       "root privileges on behalf of '%s': %s",
 			       cmdline, user, strerror(errno));
-			pipe_conv_exec_fail(data);
+			if (data->output != NULL)
+				pipe_conv_exec_fail(data);
+			g_strfreev(environ);
+			environ = environ_save;
 			die(data, ERR_EXEC_FAILED);
 		}
 		/* We're in the parent.  Wait for the child to exit.  The child
@@ -1965,11 +2160,21 @@ wrap(const char *user, const char *program,
 			pipe_conv_exec_fail(data);
 			pam_setcred(data->pamh, PAM_DELETE_CRED);
 			pam_end(data->pamh, retval);
+#ifdef HAVE_FEXECVE
+			close(fd);
+#endif
+			g_strfreev(environ);
+			environ = environ_save;
 			fail_exit(data, retval);
 		}
 
 		pam_setcred(data->pamh, PAM_DELETE_CRED);
 		pam_end(data->pamh, PAM_SUCCESS);
+#ifdef HAVE_FEXECVE
+		close(fd);
+#endif
+		g_strfreev(environ);
+		environ = environ_save;
 		if (WIFEXITED(status))
 			exit(WEXITSTATUS(status));
 		if (WIFSIGNALED(status))
@@ -1978,6 +2183,7 @@ wrap(const char *user, const char *program,
 	} else {
 		const char *cmdline;
 
+		if (env_xauthority) g_free(env_xauthority);
 		/* We're not opening a session, so we can just exec()
 		 * the program we're wrapping. */
 		pam_end(data->pamh, PAM_SUCCESS);
@@ -1988,15 +2194,23 @@ wrap(const char *user, const char *program,
 		become_super(data);
 		if (data->input != NULL) {
 			fflush(data->input);
-			fcntl(UH_INFILENO, F_SETFD, FD_CLOEXEC);
+			if (fcntl(UH_INFILENO, F_SETFD, FD_CLOEXEC) < 0)
+				g_warning("fcntl(%d, F_SETFD, FD_CLOEXEC) < 0", UH_INFILENO);
 		}
 		if (data->output != NULL) {
 			fflush(data->output);
-			fcntl(UH_OUTFILENO, F_SETFD, FD_CLOEXEC);
+			if (fcntl(UH_OUTFILENO, F_SETFD, FD_CLOEXEC) < 0)
+				g_warning("fcntl(%d, F_SETFD, FD_CLOEXEC) < 0", UH_OUTFILENO);
 		}
 		retval = pipe_conv_exec_start(data);
-		if (retval != 0)
+		if (retval != 0) {
+#ifdef HAVE_FEXECVE
+			close(fd);
+#endif
+			g_strfreev(environ);
+			environ = environ_save;
 			die(data, retval);
+		}
 #ifdef USE_STARTUP_NOTIFICATION
 		if (data->sn_id) {
 			debug_msg("userhelper: setting "
@@ -2012,11 +2226,18 @@ wrap(const char *user, const char *program,
 		       "root privileges on behalf of '%s'",
 		       cmdline, user);
 
+#ifdef HAVE_FEXECVE
+		fexecve(fd, argv + optind -1, environ);
+		close(fd);
+#else
 		execv(constructed_path, argv + optind - 1);
+#endif
 		syslog(LOG_ERR, "could not run '%s' with "
 		       "root privileges on behalf of '%s': %s",
 		       cmdline, user, strerror(errno));
 		pipe_conv_exec_fail(data);
+		g_strfreev(environ);
+		environ = environ_save;
 		die(data, ERR_EXEC_FAILED);
 	}
 }
@@ -2225,6 +2446,10 @@ main(int argc, char **argv)
 #endif
 			debug_msg("userhelper: modifying account data for %s\n",
 				  user_name);
+		} else if (optind != argc) {
+			fprintf(stderr,
+				_("Unexpected command-line arguments\n"));
+			exit(ERR_INVALID_CALL);
 		}
 
 		/* Verify that the user exists. */
