@@ -14,7 +14,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
+ * MA  02110-1301, USA.
  */
 
 /* TODO notes.
@@ -40,6 +41,7 @@
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/types.h>
+#include <sys/sysmacros.h>
 #include <errno.h>
 #include <glob.h>
 #include <libintl.h>
@@ -245,6 +247,22 @@ build_mountinfo_list(void)
 	return ret;
 }
 
+/* Free mountinfo list. */
+static void
+free_mountinfo_list(struct mountinfo *info)
+{
+	while (info != NULL) {
+		struct mountinfo *i;
+		
+		i = info;
+		info = info->next;
+		g_free (i->dir);
+		g_free (i->dev);
+		g_free (i->fstype);
+		g_free (i);
+	}
+}
+
 /* Return the mountinfo struct associated with the currently-selected row in
  * the TreeView, which is stored in the fourth column in its model. */
 static struct mountinfo *
@@ -652,26 +670,37 @@ format_one_device (const char *device)
 						 device);
                 gtk_dialog_run(GTK_DIALOG(dialog));
 	}
+	
+	free_mountinfo_list (info);
 }
 
 int
 main(int argc, char *argv[])
 {
+	char *name;
+
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	bind_textdomain_codeset(PACKAGE, "UTF-8");
 	textdomain(PACKAGE);
 	gtk_init(&argc, &argv);
 
-        if (argc > 1) {
-		char *name = strrchr (argv[0], '/');
-		
-		name = name ? name + 1: argv[0];
-		if ( !strcmp (name, "userformat") ) {
-        		format_one_device (argv[1]);
+	name = strrchr (argv[0], '/');
+	name = name ? name + 1: argv[0];
+	if (strcmp(name, "userformat") == 0) {
+		if (argc != 2) {
+			fprintf(stderr,
+				_("Unexpected command-line arguments\n"));
+			return 1;
 		}
+		format_one_device(argv[1]);
         } else {
 		gboolean run_main;
 
+		if (argc != 1) {
+			fprintf(stderr,
+				_("Unexpected command-line arguments\n"));
+			return 1;
+		}
 		run_main = create_usermount_window();
 		if (run_main) {
 			gtk_main();
